/*
 * Licensed to Elastic Search and Shay Banon under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. Elastic Search licenses this
 * file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.elasticsearch.cluster.action.index;

import com.google.inject.Inject;
import org.elasticsearch.ElasticSearchException;
import org.elasticsearch.cluster.ClusterService;
import org.elasticsearch.cluster.node.Nodes;
import org.elasticsearch.threadpool.ThreadPool;
import org.elasticsearch.transport.BaseTransportRequestHandler;
import org.elasticsearch.transport.TransportChannel;
import org.elasticsearch.transport.TransportService;
import org.elasticsearch.transport.VoidTransportResponseHandler;
import org.elasticsearch.util.component.AbstractComponent;
import org.elasticsearch.util.io.Streamable;
import org.elasticsearch.util.io.VoidStreamable;
import org.elasticsearch.util.settings.Settings;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * @author kimchy (Shay Banon)
 */
public class NodeMappingCreatedAction extends AbstractComponent {

    private final ThreadPool threadPool;

    private final TransportService transportService;

    private final ClusterService clusterService;

    private final List<Listener> listeners = new CopyOnWriteArrayList<Listener>();

    @Inject public NodeMappingCreatedAction(Settings settings, ThreadPool threadPool, TransportService transportService, ClusterService clusterService) {
        super(settings);
        this.threadPool = threadPool;
        this.transportService = transportService;
        this.clusterService = clusterService;
        transportService.registerHandler(NodeMappingCreatedTransportHandler.ACTION, new NodeMappingCreatedTransportHandler());
    }

    public void add(Listener listener) {
        listeners.add(listener);
    }

    public void remove(Listener listener) {
        listeners.remove(listener);
    }

    public void nodeMappingCreated(final NodeMappingCreatedResponse response) throws ElasticSearchException {
        Nodes nodes = clusterService.state().nodes();
        if (nodes.localNodeMaster()) {
            threadPool.execute(new Runnable() {
                @Override public void run() {
                    innerNodeIndexCreated(response);
                }
            });
        } else {
            transportService.sendRequest(clusterService.state().nodes().masterNode(),
                    NodeMappingCreatedTransportHandler.ACTION, response, VoidTransportResponseHandler.INSTANCE);
        }
    }

    private void innerNodeIndexCreated(NodeMappingCreatedResponse response) {
        for (Listener listener : listeners) {
            listener.onNodeMappingCreated(response);
        }
    }


    public static interface Listener {
        void onNodeMappingCreated(NodeMappingCreatedResponse response);
    }

    private class NodeMappingCreatedTransportHandler extends BaseTransportRequestHandler<NodeMappingCreatedResponse> {

        static final String ACTION = "cluster/nodeMappingCreated";

        @Override public NodeMappingCreatedResponse newInstance() {
            return new NodeMappingCreatedResponse();
        }

        @Override public void messageReceived(NodeMappingCreatedResponse response, TransportChannel channel) throws Exception {
            innerNodeIndexCreated(response);
            channel.sendResponse(VoidStreamable.INSTANCE);
        }
    }

    public static class NodeMappingCreatedResponse implements Streamable {

        private String index;

        private String type;

        private String nodeId;

        private NodeMappingCreatedResponse() {
        }

        public NodeMappingCreatedResponse(String index, String type, String nodeId) {
            this.index = index;
            this.type = type;
            this.nodeId = nodeId;
        }

        public String index() {
            return index;
        }

        public String type() {
            return type;
        }

        public String nodeId() {
            return nodeId;
        }

        @Override public void writeTo(DataOutput out) throws IOException {
            out.writeUTF(index);
            out.writeUTF(type);
            out.writeUTF(nodeId);
        }

        @Override public void readFrom(DataInput in) throws IOException, ClassNotFoundException {
            index = in.readUTF();
            type = in.readUTF();
            nodeId = in.readUTF();
        }
    }
}
