/*
 * Licensed to Elastic Search and Shay Banon under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. Elastic Search licenses this
 * file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.elasticsearch.cluster.routing;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import org.elasticsearch.util.io.Streamable;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.Serializable;
import java.util.List;
import java.util.Map;

import static com.google.common.collect.Lists.*;
import static com.google.common.collect.Maps.*;

/**
 * @author kimchy (shay.banon)
 */
public class RoutingTableValidation implements Serializable, Streamable {

    private boolean valid = true;

    private List<String> failures;

    private Map<String, List<String>> indicesFailures;

    public RoutingTableValidation() {
    }

    public boolean valid() {
        return valid;
    }

    public List<String> allFailures() {
        if (failures().isEmpty() && indicesFailures().isEmpty()) {
            return ImmutableList.of();
        }
        List<String> allFailures = newArrayList(failures());
        for (Map.Entry<String, List<String>> entry : indicesFailures().entrySet()) {
            for (String failure : entry.getValue()) {
                allFailures.add("Index [" + entry.getKey() + "]: " + failure);
            }
        }
        return allFailures;
    }

    public List<String> failures() {
        if (failures == null) {
            return ImmutableList.of();
        }
        return failures;
    }

    public Map<String, List<String>> indicesFailures() {
        if (indicesFailures == null) {
            return ImmutableMap.of();
        }
        return indicesFailures;
    }

    public List<String> indexFailures(String index) {
        if (indicesFailures == null) {
            return ImmutableList.of();
        }
        List<String> indexFailures = indicesFailures.get(index);
        if (indexFailures == null) {
            return ImmutableList.of();
        }
        return indexFailures;
    }

    public void addFailure(String failure) {
        valid = false;
        if (failures == null) {
            failures = newArrayList();
        }
        failures.add(failure);
    }

    public void addIndexFailure(String index, String failure) {
        valid = false;
        if (indicesFailures == null) {
            indicesFailures = newHashMap();
        }
        List<String> indexFailures = indicesFailures.get(index);
        if (indexFailures == null) {
            indexFailures = Lists.newArrayList();
            indicesFailures.put(index, indexFailures);
        }
        indexFailures.add(failure);
    }

    @Override public String toString() {
        return allFailures().toString();
    }

    @Override public void readFrom(DataInput in) throws IOException, ClassNotFoundException {
        valid = in.readBoolean();
        int size = in.readInt();
        if (size == 0) {
            failures = ImmutableList.of();
        } else {
            failures = Lists.newArrayListWithCapacity(size);
            for (int i = 0; i < size; i++) {
                failures.add(in.readUTF());
            }
        }
        size = in.readInt();
        if (size == 0) {
            indicesFailures = ImmutableMap.of();
        } else {
            indicesFailures = newHashMap();
            for (int i = 0; i < size; i++) {
                String index = in.readUTF();
                int size2 = in.readInt();
                List<String> indexFailures = newArrayListWithCapacity(size2);
                for (int j = 0; j < size2; j++) {
                    indexFailures.add(in.readUTF());
                }
                indicesFailures.put(index, indexFailures);
            }
        }
    }

    @Override public void writeTo(DataOutput out) throws IOException {
        out.writeBoolean(valid);
        if (failures == null) {
            out.writeInt(0);
        } else {
            out.writeInt(failures.size());
            for (String failure : failures) {
                out.writeUTF(failure);
            }
        }
        if (indicesFailures == null) {
            out.writeInt(0);
        } else {
            out.writeInt(indicesFailures.size());
            for (Map.Entry<String, List<String>> entry : indicesFailures.entrySet()) {
                out.writeUTF(entry.getKey());
                out.writeInt(entry.getValue().size());
                for (String failure : entry.getValue()) {
                    out.writeUTF(failure);
                }
            }
        }
    }
}
